"""
Negotiation tournaments module.
"""


from concurrent.futures import ProcessPoolExecutor, as_completed
from itertools import product
from os import cpu_count
from pathlib import Path
from random import random, shuffle
from typing import Any, Sequence

import matplotlib.pyplot as plt
import pandas as pd
from attr import asdict, define
from rich import print
from rich.progress import track

from negmas.helpers import unique_name
from negmas.helpers.inout import dump
from negmas.helpers.strings import shortest_unique_names
from negmas.helpers.types import get_class, get_full_type_name
from negmas.inout import Scenario
from negmas.mechanisms import Mechanism
from negmas.negotiators import Negotiator
from negmas.preferences.ops import (
    ScenarioStats,
    calc_outcome_distances,
    calc_outcome_optimality,
    calc_scenario_stats,
    estimate_max_dist,
)
from negmas.sao.mechanism import SAOMechanism
from negmas.serialization import serialize, to_flat_dict

__all__ = ["run_negotiation", "cartesian_tournament", "SimpleTournamentResults"]


@define
class SimpleTournamentResults:
    scores: pd.DataFrame
    """All scores per negotiator"""
    details: pd.DataFrame
    """All negotiation results"""
    scores_summary: pd.DataFrame
    """All score statistics summarized per type"""
    final_scores: pd.DataFrame
    """A list of negotiators and their final scores sorted from highest (winner) to lowest score"""
    path: Path | None = None
    """Location at which the logs are stored"""


def run_negotiation(
    s: Scenario,
    partners: tuple[type[Negotiator]],
    partner_names: tuple[str] | None = None,
    partner_params: tuple[dict[str, Any]] | None = None,
    rep: int = 0,
    path: Path | None = None,
    mechanism_type: type[Mechanism] = SAOMechanism,
    mechanism_params: dict[str, Any] | None = None,
    full_names: bool = True,
    verbosity: int = 0,
    plot=False,
    plot_params: dict[str, Any] | None = None,
    run_id: int | str | None = None,
    stats: ScenarioStats | None = None,
    annotation: dict[str, Any] | None = None,
    private_infos: tuple[dict[str, Any]] | None = None,
    id_reveals_type: bool = False,
    name_reveals_type: bool = True,
) -> dict[str, Any]:
    """
    Run a single negotiation with fully specified parameters

    Args:
        s: The `Scenario` representing the negotiation (outcome space and preferences).
        partners: The partners running the negotiation in order of addition to the mechanism.
        partner_names: Names of partners. Either `None` for defaults or a tuple of the same length as `partners`
        partner_params: Parameters used to create the partners. Either `None` for defaults or a tuple of the same length as `partners`
        rep: The repetition number for this run of the negotiation
        path: A folder to save the logs into. If not given, no logs will be saved.
        mechanism_type: the type of the `Mechanism` to use for this negotiation
        mechanism_params: The parameters used to create the `Mechanism` or `None` for defaults
        full_names: Use full names for partner names (only used if `partner_names` is None)
        verbosity: Verbosity level as an integer
        plot: If true, save a plot of the negotiation (only if `path` is given)
        plot_params: Parameters to pass to the plotting function
        run_id: A unique ID for this run. If not given one is generated based on date and time
        stats: statistics of the scenario. If not given or `path` is `None`, statistics are not saved
        annotation: Common information saved in the mechanism's annotation (accessible by negotiators using `self.nmi.annotation`). `None` for nothing
        private_infos: Private information saved in the negotiator's `private_info` attribute (accessible by negotiators as `self.private_info`). `None` for nothing
        id_reveals_type: Each negotiator ID will reveal its type.
        name_reveals_type: Each negotiator name will reveal its type.


    Returns:
        A dictionary of negotiation results that contains the final state of the negotiation alongside other information
    """
    if path:
        path = Path(path)
        for name in ("negotiations", "plots", "results"):
            (path / name).mkdir(exist_ok=True, parents=True)
    assert s.outcome_space is not None
    if not run_id:
        run_id = unique_name("run")
    run_id = str(run_id)
    if mechanism_params is None:
        mechanism_params = dict()
    if annotation is None:
        annotation = dict(rep=rep)
    else:
        annotation["rep"] = rep
    if partner_params is None:
        partner_params = tuple(dict() for _ in partners)
    if private_infos is None:
        private_infos = tuple(dict() for _ in partners)
    assert mechanism_params is not None
    assert all(_ is not None for _ in partner_params)

    def _name(a: Negotiator) -> str:
        name = a.short_type_name if not full_names else a.type_name
        if name is None:
            name = get_full_type_name(type(a))
        return name

    max_utils = [_.max() for _ in s.ufuns]

    mechanism_params["name"] = s.outcome_space.name
    mechanism_params["verbosity"] = verbosity - 1
    mechanism_params["annotation"] = annotation

    m = mechanism_type(outcome_space=s.outcome_space, **mechanism_params)
    complete_names, negotiators = [], []
    for type_, p, pinfo in zip(partners, partner_params, private_infos):
        negotiator = type_(**p, private_info=pinfo)
        name = _name(negotiator)
        complete_names.append(name)
        negotiators.append(negotiator)
        if p:
            name += str(hash(str(p)))
    if not partner_names:
        partner_names = tuple(complete_names)
    for l, (negotiator, name, u) in enumerate(zip(negotiators, partner_names, s.ufuns)):
        if id_reveals_type:
            negotiator.id = f"{name}@{l}"
        else:
            negotiator.id = unique_name("n", add_time=False, sep="")
        if name_reveals_type:
            negotiator.name = f"{name}@{l}"
        else:
            negotiator.name = unique_name("n", add_time=False, sep="")
        complete_names.append(name)
        m.add(negotiator, ufun=u)

    state = m.run()
    param_dump = tuple(str(to_flat_dict(_)) if _ else None for _ in partner_params)
    if all(_ is None for _ in param_dump):
        param_dump = None
    agreement_utils = tuple(u(state.agreement) for u in s.ufuns)
    reservations = tuple(u.reserved_value for u in s.ufuns)
    if verbosity > 0:
        print(
            f" {partner_names} on {s.outcome_space.name}: {state.agreement} in {state.relative_time:4.2%} of the time with advantages: {tuple(a - b for a, b in zip(agreement_utils, reservations))}, "
        )
    run_record = asdict(state)
    run_record["utilities"] = agreement_utils
    run_record["max_utils"] = max_utils
    run_record["reserved_values"] = reservations
    run_record["partners"] = partner_names
    run_record["params"] = param_dump
    run_record["scenario"] = m.name
    run_record["run_id"] = run_id
    run_record["negotiator_names"] = m.negotiator_names
    run_record["negotiator_ids"] = m.negotiator_ids
    run_record["negotiator_types"] = [_.type_name for _ in m.negotiators]

    if m.nmi.annotation:
        run_record.update(m.nmi.annotation)
    if stats is not None:
        dists = calc_outcome_distances(agreement_utils, stats)
        run_record.update(
            to_flat_dict(
                calc_outcome_optimality(dists, stats, estimate_max_dist(s.ufuns))
            )
        )

    file_name = f"{s.outcome_space.name}_{'_'.join(partner_names)}_{rep}"
    if path:

        def save_as_df(data: list[tuple], names, file_name):
            pd.DataFrame(data=data, columns=names).to_csv(file_name, index=False)

        full_name = path / "negotiations" / f"{file_name}.csv"
        assert not full_name.exists(), f"{full_name} already found"
        if issubclass(mechanism_type, SAOMechanism):
            save_as_df(m.extended_trace, ("step", "negotiator", "offer"), full_name)  # type: ignore
            for i, negotiator in enumerate(m.negotiators):
                neg_name = (
                    path
                    / "negotiator_behavior"
                    / file_name
                    / f"{negotiator.name}_at{i}.csv"
                )
                assert not neg_name.exists(), f"{neg_name} already found"
                neg_name.parent.mkdir(parents=True, exist_ok=True)
                save_as_df(m.negotiator_full_trace(negotiator.id), ("time", "relative_time", "step", "offer", "response"), neg_name)  # type: ignore
        else:
            pd.DataFrame.from_records(serialize(m.history)).to_csv(
                full_name, index=False
            )
        full_name = path / "results" / f"{file_name}.json"
        assert not full_name.exists(), f"{full_name} already found"
        dump(run_record, full_name)
        if plot:
            if plot_params is None:
                plot_params = dict()
            plot_params["save_fig"] = (True,)
            full_name = path / "plots" / f"{file_name}.png"
            m.plot(path=path, fig_name=full_name, **plot_params)
            try:
                plt.close(plt.gcf())
            except:
                pass
    return run_record


def cartesian_tournament(
    competitors: list[type[Negotiator] | str] | tuple[type[Negotiator] | str, ...],
    scenarios: list[Scenario] | tuple[Scenario, ...],
    competitor_params: Sequence[dict | None] | None = None,
    rotate_ufuns: bool = True,
    n_repetitions: int = 1,
    path: Path | None = None,
    njobs: int = 0,
    mechanism_type: type[Mechanism] = SAOMechanism,
    mechanism_params: dict[str, Any] | None = None,
    # full_names: bool = True,
    plot_fraction: float = 0.0,
    verbosity: int = 1,
    self_play: bool = True,
    randomize_runs: bool = True,
    save_every: int = 0,
    save_stats: bool = True,
    final_score: tuple[str, str] = ("advantage", "mean"),
    id_reveals_type: bool = False,
    name_reveals_type: bool = True,
    shorten_names: bool = True,
) -> SimpleTournamentResults:
    """A simplified version of Cartesian tournaments not using the internal machinay of NegMAS  tournaments

    Args:
        competitors: A tuple of the competing negotiator types.
        scenarios: A tuple of base scenarios to use for the tournament.
        competitor_params: Either None for no-parameters or a tuple of dictionaries with parameters to initialize the competitors (in order).
        rotate_ufuns: If `True`, the ufuns will be rotated over negotiator positions (for bilateral negotiation this leads to two scenarios for each input scenario with reversed ufun order).
        n_repetitions: Number of times to repeat each scenario/partner combination
        path: Path on disk to save the results and details of this tournament. Pass None to disable logging
        n_jobs: Number of parallel jobs to run. -1 means running serially (useful for debugging) and 0 means using all cores.
        mechanism_type: The mechanism (protocol) used for all negotiations.
        mechanism_params: Parameters of the mechanism (protocol). Usually you need to pass one or more of the following:
                          time_limit (in seconds), n_steps (in rounds), p_ending (probability of ending the negotiation every step).
        plot_fraction: fraction of negotiations for which plots are to be saved (only if `path` is not `None`)
        verbosity: Verbosity level (minimum is 0)
        self_play: Allow negotiations in which all partners are of the same type
        randomize_runs: If `True` negotiations will be run in random order, otherwise each scenario/partner combination will be finished before starting on the next
        save_every: Number of negotiations after which we dump details and scores
        save_stats: Whether to calculate and save extra statistics like pareto_optimality, nash_optimality, kalai_optimality, etc
        final_score: A tuple of two strings giving the metric used for ordering the negotiators for the final score:
                     First string can be one of the following (advantage, utility,
                     partner_welfare, welfare) or any statistic from the set calculated if `save_stats` is `True`.
                     The second string can be mean, median, min, max, or std. The default is ('advantage', 'mean')
        id_reveals_type: Each negotiator ID will reveal its type.
        name_reveals_type: Each negotiator name will reveal its type.
        shorten_names: If True, shorter versions of names will be used for results

    Returns:
        A pandas DataFrame with all negotiation results.
    """
    competitors = [get_class(_) for _ in competitors]
    if competitor_params is None:
        competitor_params = [dict() for _ in competitors]

    def make_scores(record: dict[str, Any]) -> list[dict[str, float]]:
        utils, partners = record["utilities"], record["partners"]
        reserved_values = record["reserved_values"]
        max_utils = record["max_utils"]
        scores = []
        for i, (u, r, a, m) in enumerate(
            zip(utils, reserved_values, partners, max_utils)
        ):
            n_p = len(partners)
            bilateral = n_p == 2
            basic = dict(
                strategy=a,
                utility=u,
                reserved_value=r,
                advantage=(u - r) / (m - r),
                partner_welfare=sum(_ for j, _ in enumerate(utils) if j != i)
                / (n_p - 1),
                welfare=sum(_ for _ in utils) / n_p,
                scenario=record["scenario"],
                partners=(partners[_] for _ in range(len(partners)) if _ != i)
                if not bilateral
                else partners[1 - i],
            )
            for c in (
                "nash_optimality",
                "kalai_optimality",
                "max_welfare_optimality",
                "pareto_optimality",
            ):
                if c in record:
                    basic[c] = record[c]
            scores.append(basic)
        return scores

    runs = []
    scenarios_path = path if path is None else Path(path) / "scenarios"
    if scenarios_path is not None:
        scenarios_path.mkdir(exist_ok=True, parents=True)
    stats = None

    def shorten(name):
        #        for s in ("Negotiator", "Agent"):
        #            x = name.replace(s, "")
        #            if not x:
        #                return name
        #            name = x
        return name

    if shorten_names:
        competitor_names = [
            shorten(_)
            for _ in shortest_unique_names([get_full_type_name(_) for _ in competitors])
        ]
    else:
        competitor_names = [get_full_type_name(_) for _ in competitors]
    competitor_info = list(
        zip(competitors, competitor_params, competitor_names, strict=True)
    )
    for s in scenarios:
        n = len(s.ufuns)
        partners_list = list(product(*tuple([competitor_info] * n)))
        if not self_play:
            partners_list = [_ for _ in partners_list if len(set(serialize(_))) > 1]

        ufun_sets = [list(s.ufuns)]
        for i, u in enumerate(s.ufuns):
            u.name = f"{i}_{u.name}"
        if rotate_ufuns:
            for _ in range(len(ufun_sets)):
                ufuns = ufun_sets[-1]
                ufun_sets.append([ufuns[-1]] + ufuns[:-1])
        original_name = s.outcome_space.name
        # original_ufun_names = [_.name for _ in s.ufuns]
        for i, ufuns in enumerate(ufun_sets):
            if len(ufun_sets) > 1:
                for j, u in enumerate(ufuns):
                    n = "_".join(u.name.split("_")[1:])
                    u.name = f"{j}_{n}"
                scenario = Scenario(
                    type(s.outcome_space)(
                        issues=s.outcome_space.issues,
                        name=f"{original_name}_{i}" if i else original_name,
                    ),
                    tuple(ufuns),
                )
            else:
                scenario = s
            this_path = None
            if scenarios_path:
                this_path = scenarios_path / str(scenario.outcome_space.name)
                scenario.to_yaml(this_path)
            if save_stats:
                stats = calc_scenario_stats(scenario.ufuns)
                if this_path:
                    dump(serialize(stats), this_path / "stats.json")
            for partners in partners_list:
                runs += [
                    dict(
                        s=scenario,
                        partners=[_[0] for _ in partners],
                        partner_names=[_[2] for _ in partners],
                        partner_params=[_[1] for _ in partners],
                        rep=i,
                        annotation=dict(rep=i, n_repetitions=n_repetitions),
                        path=path if path else None,
                        mechanism_type=mechanism_type,
                        mechanism_params=mechanism_params,
                        full_names=True,
                        verbosity=verbosity - 1,
                        plot=random() < plot_fraction,
                        stats=stats,
                        id_reveals_type=id_reveals_type,
                        name_reveals_type=name_reveals_type,
                    )
                    for i in range(n_repetitions)
                ]
    if randomize_runs:
        shuffle(runs)
    if verbosity > 0:
        print(
            f"Will run {len(runs)} negotiations on {len(scenarios)} scenarios between {len(competitors)} competitors",
            flush=True,
        )
    results, scores = [], []
    results_path = path if not path else path / "details.csv"
    scores_path = path if not path else path / "all_scores.csv"

    def process_record(record, results=results, scores=scores):
        results.append(record)
        scores += make_scores(record)
        if results_path and save_every and i % save_every == 0:
            pd.DataFrame.from_records(results).to_csv(results_path, index_label="index")
            pd.DataFrame.from_records(scores).to_csv(scores_path, index_label="index")
        return results, scores

    def get_run_id(info):
        return hash(str(serialize(info)))

    if njobs < 0:
        for i, info in enumerate(
            track(runs, total=len(runs), description="Negotiations")
        ):
            process_record(run_negotiation(**info, run_id=get_run_id(info)))

    else:
        futures = []
        n_cores = cpu_count()
        if n_cores is None:
            n_cores = 4
        cpus = min(n_cores, njobs) if njobs else cpu_count()
        with ProcessPoolExecutor(max_workers=cpus) as pool:
            for info in runs:
                futures.append(
                    pool.submit(run_negotiation, **info, run_id=get_run_id(info))
                )
            for i, f in enumerate(
                track(
                    as_completed(futures),
                    total=len(futures),
                    description="Negotiations",
                )
            ):
                process_record(f.result())

    scores_df = pd.DataFrame.from_records(scores)
    if scores_path is not None:
        scores_df.to_csv(scores_path, index_label="index")
    type_scores = (
        scores_df[[_ for _ in scores_df.columns if _ not in ("scenario", "partners")]]
        .groupby("strategy")
        .describe()
        # .agg(["min", "mean", "std", "median", "max"])
        .sort_values(final_score, ascending=False)
    )
    final = pd.DataFrame()
    if type_scores is not None:
        if scores_path:
            type_scores.to_csv(scores_path.parent / "type_scores.csv")
        final = type_scores[final_score]
        final.name = "score"
        final = final.reset_index()
        if scores_path:
            final.to_csv(scores_path.parent / "scores.csv", index_label="index")
        if verbosity > 0:
            print(final)
    details_df = pd.DataFrame.from_records(results)
    if results_path:
        details_df.to_csv(results_path, index_label="index")
    return SimpleTournamentResults(
        scores=scores_df,
        details=details_df,
        scores_summary=type_scores,
        final_scores=final,
        path=path,
    )


if __name__ == "__main__":
    ...



from __future__ import annotations

from typing import TYPE_CHECKING

from ...components import AcceptancePolicy, Model, OfferingPolicy
from .mapneg import MAPNegotiator

if TYPE_CHECKING:
    from ...components import GBComponent
    from ...negotiators.base import GBNegotiator

__all__ = ["make_boa", "BOANegotiator"]


def make_boa(
    acceptance: AcceptancePolicy | GBNegotiator | None = None,
    offering: OfferingPolicy | GBNegotiator | None = None,
    model: Model | None = None,
    extra_components: list[GBComponent] | None = None,
    extra_component_names: list[str] | None = None,
    **kwargs,
):
    """
    A negotiator that is  constructed from three components:

    1. A `PartnerUFunModel` to learn the ufun of the partner
    2. An `AcceptanceStrategy` That is used for responding to offers.
    3. An `OfferingStrategy` That is used for generating offers.


    For all callbacks, partner-model is called first, followed by the acceptance strategy followed by the offering strategy

    """

    return MAPNegotiator(
        acceptance=acceptance,
        offering=offering,
        models=[model] if model else None,
        model_names=["model"] if model else None,
        extra_components=extra_components,
        extra_component_names=extra_component_names,
        type_name="BOANegotiator",
        **kwargs,
    )


class BOANegotiator(MAPNegotiator):
    """
    A negotiator that is  constructed from three components:

    1. A `PartnerUFunModel` to learn the ufun of the partner
    2. An `AcceptanceStrategy` That is used for responding to offers.
    3. An `OfferingStrategy` That is used for generating offers.


    For all callbacks, partner-model is called first, followed by the acceptance strategy followed by the offering strategy

    """

    def __init__(
        *args,
        acceptance: AcceptancePolicy | GBNegotiator | None = None,
        offering: OfferingPolicy | GBNegotiator | None = None,
        model: Model | None = None,
        extra_components: list[GBComponent] | None = None,
        extra_component_names: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(
            acceptance=acceptance,
            offering=offering,
            models=[model] if model else None,
            model_names=["model"] if model else None,
            extra_components=extra_components,
            extra_component_names=extra_component_names,
            type_name="BOANegotiator",
            **kwargs,
        )



from __future__ import annotations

import math
import sys
from functools import partial

import click

try:
    from PyQt5 import QtCore, QtGui, QtWidgets
except:
    pass
else:
    MAX_ROWS = 21

    try:
        import qdarkstyle

        _has_qdarkstyle = True
    except ModuleNotFoundError:
        _has_qdarkstyle = False

    _GTypeRole = QtCore.Qt.UserRole
    _missing = object()

    class GStyle:
        _base_style = """
            ._OptionLabel {
                font-size: 16px;
                font: bold;
                font-family: arial;
                }
            ._HelpLabel {
                font-family: arial;
                font-size: 14px;
                }
            ._InputComboBox{
                font-size: 16px;
                }
            ._InputLineEdit{
                font-size: 16px;
                }
            ._InputCheckBox{
                font-size: 16px;
                }
            ._InputSpinBox{
                font-size: 16px;
                }
            ._InputTabWidget{
                font: bold;
                font-size: 16px;
                }
            .GListView{
                font-size: 16px;
                }
            QToolTip{
                font-family: arial;
                }
            """

        def __init__(self, style=""):
            if not GStyle.check_style(style):
                self.text_color = "black"
                self.placehoder_color = "#898b8d"
                self.stylesheet = (
                    GStyle._base_style
                    + """
                        ._Spliter{
                            border: 1px inset gray;
                            }
                        """
                )
            elif style == "qdarkstyle":
                self.text_color = "#eff0f1"
                self.placehoder_color = "#898b8d"
                self.stylesheet = (
                    qdarkstyle.load_stylesheet_pyqt5()
                    + GStyle._base_style
                    + """
                        .GListView{
                            padding: 5px;
                            }
                        ._Spliter{
                            border: 5px solid gray;
                            }
                        """
                )

        @staticmethod
        def check_style(style):
            if style == "qdarkstyle":
                return _has_qdarkstyle
            return False

    _gstyle = GStyle()

    class GListView(QtWidgets.QListView):
        def __init__(self, opt):
            super().__init__()
            self.nargs = opt.nargs
            self.model = GItemModel(
                opt.nargs, parent=self, opt_type=opt.type, default=opt.default
            )
            self.setModel(self.model)
            self.delegate = GEditDelegate(self)
            self.setItemDelegate(self.delegate)
            self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
            if self.nargs == -1:
                self.keyPressEvent = self.key_press
                self.setToolTip(
                    "'a': add a new item blow the selected one\n"
                    "'d': delete the selected item"
                )

        def key_press(self, e):
            if self.nargs == -1:
                if e.key() == QtCore.Qt.Key_A:
                    if len(self.selectedIndexes()) == 0:
                        self.model.insertRow(0)
                    else:
                        for i in self.selectedIndexes():
                            self.model.insertRow(i.row() + 1)
                if e.key() == QtCore.Qt.Key_D:
                    si = self.selectedIndexes()
                    for i in si:
                        self.model.removeRow(i.row())
            super().keyPressEvent(e)

    class GItemModel(QtGui.QStandardItemModel):
        def __init__(self, n, parent=None, opt_type=click.STRING, default=None):
            super(QtGui.QStandardItemModel, self).__init__(0, 1, parent)
            self.type = opt_type
            for row in range(n):
                if hasattr(default, "__len__"):
                    self.insertRow(row, default[row])
                else:
                    self.insertRow(row, default)

        def insertRow(self, idx, val=""):
            super().insertRow(idx)

            index = self.index(idx, 0, QtCore.QModelIndex())
            if val is None or val == "":
                self.setData(
                    index,
                    QtGui.QBrush(QtGui.QColor(_gstyle.placehoder_color)),
                    role=QtCore.Qt.ForegroundRole,
                )
            else:
                self.setData(index, val)

        def data(self, index, role=QtCore.Qt.DisplayRole):
            if role == QtCore.Qt.DisplayRole:
                dstr = QtGui.QStandardItemModel.data(self, index, role)
                if dstr == "" or dstr is None:
                    if isinstance(self.type, click.types.Tuple):
                        row = index.row()
                        if 0 <= row < len(self.type.types):
                            tp = self.type.types[row]
                            dstr = tp.name
                    else:
                        dstr = self.type.name
                    return dstr

            if role == _GTypeRole:
                tp = click.STRING
                if isinstance(self.type, click.types.Tuple):
                    row = index.row()
                    if 0 <= row < len(self.type.types):
                        tp = self.type.types[row]
                elif isinstance(self.type, click.types.ParamType):
                    tp = self.type
                return tp

            return QtGui.QStandardItemModel.data(self, index, role)

    class GEditDelegate(QtWidgets.QStyledItemDelegate):
        def createEditor(self, parent, option, index):
            tp = index.data(role=_GTypeRole)
            if isinstance(tp, click.Path):
                led = GLineEdit_path.from_option(tp, parent)
            else:
                led = QtWidgets.QLineEdit(parent)
            led.setPlaceholderText(tp.name)
            led.setValidator(select_type_validator(tp))
            return led

        def setEditorData(self, editor, index):
            item_var = index.data(role=QtCore.Qt.EditRole)
            if item_var is not None:
                editor.setText(str(item_var))

        def setModelData(self, editor, model, index):
            data_str = editor.text()
            if data_str == "" or data_str is None:
                model.setData(
                    index,
                    QtGui.QBrush(QtGui.QColor(_gstyle.placehoder_color)),
                    role=QtCore.Qt.ForegroundRole,
                )
            else:
                model.setData(
                    index,
                    QtGui.QBrush(QtGui.QColor(_gstyle.text_color)),
                    role=QtCore.Qt.ForegroundRole,
                )
            QtWidgets.QStyledItemDelegate.setModelData(self, editor, model, index)

    def generate_label(opt):
        show_name = getattr(opt, "show_name", _missing)
        show_name = opt.name if show_name is _missing else show_name
        param = _OptionLabel(show_name)
        param.setToolTip(getattr(opt, "help", None))
        return param

    class GStringLineEditor(click.types.StringParamType):
        def to_widget(self, opt, validator=None):
            value = _InputLineEdit()
            value.setPlaceholderText(self.name)
            if opt.default:
                value.setText(str(opt.default))
            if getattr(opt, "hide_input", False):
                value.setEchoMode(QtWidgets.QLineEdit.Password)
            value.setValidator(validator)

            def to_command():
                return [opt.opts[0], value.text()]

            return [generate_label(opt), value], to_command

    class GIntLineEditor(GStringLineEditor):
        def to_widget(self, opt):
            return GStringLineEditor.to_widget(
                self, opt, validator=QtGui.QIntValidator()
            )

    class GFloatLineEditor(GStringLineEditor):
        def to_widget(self, opt):
            return GStringLineEditor.to_widget(
                self, opt, validator=QtGui.QDoubleValidator()
            )

    class GFileDialog(QtWidgets.QFileDialog):
        def __init__(
            self, *args, exists=False, file_okay=True, dir_okay=True, **kwargs
        ):
            super().__init__(*args, **kwargs)
            self.setOption(QtWidgets.QFileDialog.DontUseNativeDialog, True)
            self.setLabelText(QtWidgets.QFileDialog.Accept, "Select")
            if (exists, file_okay, dir_okay) == (True, True, False):
                self.setFileMode(QtWidgets.QFileDialog.ExistingFile)
            elif (exists, file_okay, dir_okay) == (False, True, False):
                self.setFileMode(QtWidgets.QFileDialog.AnyFile)
            elif (exists, file_okay, dir_okay) == (True, False, True):
                self.setFileMode(QtWidgets.QFileDialog.Directory)
            elif (exists, file_okay, dir_okay) == (False, False, True):
                self.setFileMode(QtWidgets.QFileDialog.Directory)
            elif exists == True:
                self.setFileMode(QtWidgets.QFileDialog.ExistingFile)
                self.accept = self.accept_all
            elif exists == False:
                self.setFileMode(QtWidgets.QFileDialog.AnyFile)
                self.accept = self.accept_all

        def accept_all(self):
            super().done(QtWidgets.QFileDialog.Accepted)

    class GLineEdit_path(QtWidgets.QLineEdit):
        def __init__(self, parent=None, exists=False, file_okay=True, dir_okay=True):
            super().__init__(parent)
            self.action = self.addAction(
                self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon),
                QtWidgets.QLineEdit.TrailingPosition,
            )
            self.fdlg = GFileDialog(
                self,
                "Select File Dialog",
                "./",
                "*",
                exists=exists,
                file_okay=file_okay,
                dir_okay=dir_okay,
            )
            self.action.triggered.connect(self.run_dialog)

        def run_dialog(self):
            if self.fdlg.exec() == QtWidgets.QFileDialog.Accepted:
                self.setText(self.fdlg.selectedFiles()[0])

        @staticmethod
        def from_option(opt, parent=None):
            return GLineEdit_path(
                parent=parent,
                exists=opt.exists,
                file_okay=opt.file_okay,
                dir_okay=opt.dir_okay,
            )

    class GPathGLindEidt_path(click.types.Path):
        def to_widget(self, opt):
            value = GLineEdit_path(
                exists=self.exists, file_okay=self.file_okay, dir_okay=self.dir_okay
            )
            value.setPlaceholderText(self.name)
            if opt.default:
                value.setText(str(opt.default))

            def to_command():
                return [opt.opts[0], value.text()]

            return [generate_label(opt), value], to_command

    class _GLabeledSlider(QtWidgets.QSlider):
        def __init__(self, min, max, val):
            super().__init__(QtCore.Qt.Horizontal)
            self.min, self.max = min, max

            self.setMinimum(min)
            self.setMaximum(max)
            self.setValue(val)

            self.label = self.__init_label()

        def __init_label(self):
            l = max(
                math.ceil(math.log10(abs(x))) if x != 0 else 1
                for x in [self.min, self.max]
            )
            l += 1
            return QtWidgets.QLabel("0" * l)

    def argument_command(to_command):
        def tc():
            a = to_command()
            return a[1:]

        return tc

    class GSlider(QtWidgets.QHBoxLayout):
        def __init__(self, min=0, max=10, default=None, *args, **kwargs):
            super(QtWidgets.QHBoxLayout, self).__init__()

            self.min, self.max, self.default = min, max, default
            self.label = self.__init_label()
            self.slider = self.__init_slider()

            self.label.setText(str(self.default))

            self.addWidget(self.slider)
            self.addWidget(self.label)

        def value(self):
            return self.slider.value()

        def __init_slider(self):
            slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
            slider.setMinimum(self.min)
            slider.setMaximum(self.max)
            default_val = (self.min + self.max) // 2
            if isinstance(self.default, int):
                if self.min <= self.default <= self.max:
                    default_val = self.default
            self.default = default_val
            slider.setValue(default_val)
            slider.valueChanged.connect(lambda x: self.label.setText(str(x)))
            return slider

        def __init_label(self):
            l = max(
                math.ceil(math.log10(abs(x))) if x != 0 else 1
                for x in [self.min, self.max]
            )
            l += 1
            return QtWidgets.QLabel("0" * l)

    class GIntRangeGSlider(click.types.IntRange):
        def to_widget(self, opt):
            value = GSlider(min=self.min, max=self.max, default=opt.default)

            def to_command():
                return [opt.opts[0], str(value.value())]

            return [generate_label(opt), value], to_command

    class GIntRangeSlider(click.types.IntRange):
        def to_widget(self, opt):
            value = QtWidgets.QSlider(QtCore.Qt.Horizontal)
            value.setMinimum(self.min)
            value.setMaximum(self.max)

            default_val = (self.min + self.max) // 2
            if isinstance(opt.default, int):
                if self.min <= opt.default <= self.max:
                    default_val = opt.default
            value.setValue(default_val)

            def to_command():
                return [opt.opts[0], str(value.value())]

            return [generate_label(opt), value], to_command

    class GIntRangeLineEditor(click.types.IntRange):
        def to_widget(self, opt):
            value = QtWidgets.QLineEdit()
            # TODO: set validator

            def to_command():
                return [opt.opts[0], value.text()]

            return [generate_label(opt), value], to_command

    def bool_flag_option(opt):
        checkbox = _InputCheckBox(opt.name)
        if opt.default:
            checkbox.setCheckState(2)
        # set tip
        checkbox.setToolTip(opt.help)

        def to_command():
            if checkbox.checkState():
                return [opt.opts[0]]
            else:
                return opt.secondary_opts

        return [checkbox], to_command

    class GChoiceComboBox(click.types.Choice):
        def to_widget(self, opt):
            cb = _InputComboBox()
            cb.addItems(self.choices)

            def to_command():
                return [opt.opts[0], cb.currentText()]

            return [generate_label(opt), cb], to_command

    def count_option(opt):
        sb = _InputSpinBox()

        def to_command():
            return [opt.opts[0]] * int(sb.text())

        return [generate_label(opt), sb], to_command

    class GTupleGListView(click.Tuple):
        def to_widget(self, opt):
            view = GListView(opt)

            def to_command():
                _ = [opt.opts[0]]
                for idx in range(view.model.rowCount()):
                    _.append(view.model.item(idx).text())
                return _

            return [generate_label(opt), view], to_command

    def multi_text_arguement(opt):
        value = GListView(opt)

        def to_command():
            _ = []
            for idx in range(value.model.rowCount()):
                _.append(value.model.item(idx).text())
            # if opt.required and value.model.rowCount() == 0:
            # raise click.exceptions.BadParameter("Required")
            # print(opt.__dict__)
            return _

        # return [QtWidgets.QLabel(opt.name), value], to_command
        return [_OptionLabel(opt.name), value], to_command

    def select_type_validator(tp: click.types.ParamType) -> QtGui.QValidator:
        """select the right validator for `tp`"""

        if isinstance(tp, click.types.IntParamType):
            return QtGui.QIntValidator()
        elif isinstance(tp, click.types.FloatParamType):
            return QtGui.QDoubleValidator()

        return None

    def select_opt_validator(opt):
        """select the right validator for `opt`"""

        return select_type_validator(opt.type)

    def opt_to_widget(opt):
        if opt.nargs > 1:
            return GTupleGListView.to_widget(opt.type, opt)
        elif getattr(opt, "is_bool_flag", False):
            return bool_flag_option(opt)
        elif getattr(opt, "count", False):
            return count_option(opt)
        elif isinstance(opt.type, click.types.Choice):
            return GChoiceComboBox.to_widget(opt.type, opt)
        elif isinstance(opt.type, click.types.Path):
            return GPathGLindEidt_path.to_widget(opt.type, opt)
        elif isinstance(opt.type, click.types.IntRange):
            return GIntRangeGSlider.to_widget(opt.type, opt)
        elif isinstance(opt.type, click.types.IntParamType):
            return GIntLineEditor.to_widget(opt.type, opt)
        elif isinstance(opt.type, click.types.FloatParamType):
            return GFloatLineEditor.to_widget(opt.type, opt)
        else:
            return GStringLineEditor.to_widget(opt.type, opt)

    def _to_widget(opt):
        # customed widget

        if isinstance(opt.type, click.types.FuncParamType):
            if hasattr(opt.type.func, "to_widget"):
                return opt.type.func.to_widget()
        elif hasattr(opt.type, "to_widget"):
            return opt.type.to_widget()

        if isinstance(opt, click.core.Argument):
            if opt.nargs == 1:
                w, tc = opt_to_widget(opt)

                return w, argument_command(tc)
            elif opt.nargs > 1 or opt.nargs == -1:
                return multi_text_arguement(opt)
        else:
            return opt_to_widget(opt)

    def layout_append_opts(layout, opts):
        params_func = []
        widgets = []
        i = 0

        for i, para in enumerate(opts):
            widget, value_func = _to_widget(para)
            widgets.append(widget)
            params_func.append(value_func)

            for idx, w in enumerate(widget):
                if isinstance(w, QtWidgets.QLayout):
                    layout.addLayout(w, i, idx)
                else:
                    layout.addWidget(w, i, idx)

        return layout, params_func, widgets

    def generate_sysargv(cmd_list):
        argv_list = []

        for name, func_list in cmd_list:
            argv_list.append(name)

            for value_func in func_list:
                argv_list += value_func()

        return argv_list

    class _Spliter(QtWidgets.QFrame):
        def __init__(self, parent=None):
            super().__init__(parent=parent)
            self.setFrameShape(QtWidgets.QFrame.HLine)

    class _InputComboBox(QtWidgets.QComboBox):
        pass

    class _InputTabWidget(QtWidgets.QTabWidget):
        pass

    class _HelpLabel(QtWidgets.QLabel):
        pass

    class _OptionLabel(QtWidgets.QLabel):
        pass

    class _InputLineEdit(QtWidgets.QLineEdit):
        pass

    class _InputCheckBox(QtWidgets.QCheckBox):
        pass

    class _InputSpinBox(QtWidgets.QSpinBox):
        pass

    class CommandLayout(QtWidgets.QGridLayout):
        def __init__(self, func, run_exit, parent_layout=None):
            super().__init__()
            self.parent_layout = parent_layout
            self.func = func
            self.run_exit = run_exit

            if func.help:
                label = _HelpLabel(func.help)
                label.setWordWrap(True)
                self.addWidget(label, 0, 0, 1, 2)
                frame = _Spliter()
                self.addWidget(frame, 1, 0, 1, 4)
            self.params_func, self.widgets = self.append_opts(self.func.params)

        def add_sysargv(self):
            if hasattr(self.parent_layout, "add_sysargv"):
                self.parent_layout.add_sysargv()
            sys.argv += generate_sysargv([(self.func.name, self.params_func)])

        def append_opts(self, opts):
            params_func = []
            widgets = []

            def _nwidgets(x):
                return -len(_to_widget(x)[0])

            opts = sorted(opts, key=_nwidgets)
            row, col = 0, 0
            for para in opts:
                widget, value_func = _to_widget(para)
                widgets.append(widget)
                params_func.append(value_func)

                for idx, w in enumerate(widget):
                    if isinstance(w, QtWidgets.QLayout):
                        self.addLayout(w, row + 2, 2 * col + idx)
                    else:
                        self.addWidget(w, row + 2, 2 * col + idx)
                self.setRowStretch(row, 5)
                row += 1
                if row >= MAX_ROWS - 1:
                    row = 0
                    col += 1

            return params_func, widgets

        def generate_cmd_button(self, label, cmd_slot, tooltip=""):
            button = QtWidgets.QPushButton(label)
            button.setToolTip(tooltip)
            button.clicked.connect(self.clean_sysargv)
            button.clicked.connect(self.add_sysargv)
            button.clicked.connect(cmd_slot)

            return button

        def add_cmd_button(self, label, cmd_slot, pos=None):
            run_button = self.generate_cmd_button(label, cmd_slot)

            if pos is None:
                # pos = self.rowCount() + 1, 0
                pos = 0, 2
            self.addWidget(run_button, pos[0], pos[1])

        def add_cmd_buttons(self, args):
            # row = self.rowCount() + 1
            row = 0
            cmd_layout = QtWidgets.QGridLayout()
            cmd_layout.setHorizontalSpacing(20)

            for col, arg in enumerate(args):
                button = self.generate_cmd_button(**arg)
                cmd_layout.addWidget(button, col, 0)
            self.addLayout(cmd_layout, row, 2, 1, 2)

        @QtCore.pyqtSlot()
        def clean_sysargv(self):
            sys.argv = []

    class RunCommand(QtCore.QRunnable):
        def __init__(self, func, run_exit):
            super().__init__()
            self.func = func
            self.run_exit = run_exit

        @QtCore.pyqtSlot()
        def run(self):
            print(sys.argv)
            try:
                self.func(standalone_mode=self.run_exit)
            except click.exceptions.BadParameter as bpe:
                # warning message
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Warning)
                msg.setText(bpe.format_message())
                msg.exec_()
            except Exception as bpe:
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Warning)
                msg.setText(repr(bpe))
                msg.exec_()
            # if self.outputEdit is not None:
            # self.outputEdit.show()

    class GCommand(click.Command):
        def __init__(self, new_thread=True, *arg, **args):
            super().__init__(*arg, **args)
            self.new_thread = new_thread

    class GOption(click.Option):
        def __init__(self, *arg, show_name=_missing, **args):
            super().__init__(*arg, **args)
            self.show_name = show_name

    # def normalOutputWritten(t):
    # """Append text to the QTextEdit."""
    # Maybe QTextEdit.append() works as well, but this is how I do it:
    # cursor = text.textCursor()
    # cursor.movePosition(QtGui.QTextCursor.End)
    # cursor.insertText(t)
    # text.setTextCursor(cursor)
    # text.ensureCursorVisible()

    class GuiStream(QtCore.QObject):
        textWritten = QtCore.pyqtSignal(str)

        def flush(self):
            pass

        def write(self, text):
            self.textWritten.emit(str(text))

    class OutputEdit(QtWidgets.QTextEdit):
        def print(self, text):
            cursor = self.textCursor()
            cursor.movePosition(QtGui.QTextCursor.End)
            cursor.insertText(text)
            self.setTextCursor(cursor)
            self.ensureCursorVisible()

    class App(QtWidgets.QWidget):
        def __init__(
            self,
            func,
            run_exit,
            new_thread,
            output="gui",
            left=10,
            top=10,
            width=800,
            height=280,
        ):
            """
            Parameters
            ----------
            output : str
                'gui': [default] redirect screen output to the gui
                'term': do nothing
            """
            super().__init__()
            self.new_thread = new_thread
            self.title = func.name
            self.func = func
            self.initUI(run_exit, QtCore.QRect(left, top, width, height))
            self.threadpool = QtCore.QThreadPool()
            self.outputEdit = self.initOutput(output)

        def initOutput(self, output):
            if output == "gui":
                sys.stdout = GuiStream()
                sys.stderr = sys.stdout
                text = OutputEdit()
                text.setReadOnly(True)
                sys.stdout.textWritten.connect(text.print)
                sys.stdout.textWritten.connect(text.show)
                # sys.stdout.textWritten.connect(text.append)
                # text.show()

                return text
            else:
                return None

        def initCommandUI(self, func, run_exit, parent_layout=None):
            opt_set = CommandLayout(func, run_exit, parent_layout=parent_layout)

            if isinstance(func, click.MultiCommand):
                tabs = _InputTabWidget()

                for cmd, f in func.commands.items():
                    sub_opt_set = self.initCommandUI(f, run_exit, parent_layout=opt_set)
                    tab = QtWidgets.QWidget()
                    tab.setLayout(sub_opt_set)
                    tabs.addTab(tab, cmd)
                opt_set.addWidget(tabs, opt_set.rowCount(), 0, 1, 4)
                # return opt_set
            elif isinstance(func, click.Command):
                new_thread = getattr(func, "new_thread", self.new_thread)
                opt_set.add_cmd_buttons(
                    args=[
                        {
                            "label": "&Run",
                            "cmd_slot": partial(self.run_cmd, new_thread=new_thread),
                            "tooltip": "run command",
                        },
                        {
                            "label": "&Copy",
                            "cmd_slot": self.copy_cmd,
                            "tooltip": "copy command to clipboard",
                        },
                    ]
                )

            return opt_set

        def initUI(self, run_exit, geometry):
            self.run_exit = run_exit
            self.setWindowTitle(self.title)
            # self.setGeometry(self.left, self.top, self.width, self.height)
            self.setGeometry(geometry)
            self.opt_set = self.initCommandUI(
                self.func,
                run_exit,
            )
            self.setLayout(self.opt_set)
            self.show()

        @QtCore.pyqtSlot()
        def copy_cmd(self):
            cb = QtWidgets.QApplication.clipboard()
            cb.clear(mode=cb.Clipboard)
            cmd_text = " ".join(sys.argv)
            cb.setText(cmd_text, mode=cb.Clipboard)

            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText(f"copy '{cmd_text}' to clipboard")
            msg.exec_()

        def run_cmd(self, new_thread):
            runcmd = RunCommand(self.func, self.run_exit)

            if new_thread:
                self.threadpool.start(runcmd)
            else:
                runcmd.run()

    def gui_it(click_func, style="qdarkstyle", **argvs) -> None:
        """
        Parameters
        ----------
        click_func
        `new_thread` is used for qt-based func, like matplotlib
        """
        global _gstyle
        _gstyle = GStyle(style)
        app = QtWidgets.QApplication(sys.argv)
        app.setStyleSheet(_gstyle.stylesheet)

        # set the default value for argvs
        argvs["run_exit"] = argvs.get("run_exit", False)
        argvs["new_thread"] = argvs.get("new_thread", False)

        _ = App(click_func, **argvs)
        sys.exit(app.exec_())

    def gui_option(f: click.core.BaseCommand) -> click.core.BaseCommand:
        """decorator for adding '--gui' option to command"""

        # TODO: add run_exit, new_thread
        def run_gui_it(ctx, param, value):
            if not value or ctx.resilient_parsing:
                return
            f.params = [p for p in f.params if not p.name == "gui"]
            gui_it(f)
            ctx.exit()

        return click.option(
            "--gui",
            is_flag=True,
            callback=run_gui_it,
            help="run with gui",
            expose_value=False,
            is_eager=False,
        )(f)



# -*- coding: utf-8 -*-
"""
This module defines the base classes for worlds within which multiple agents engage in situated negotiations


The `Agent` class encapsulates the managing entity that creates negotiators to engage in negotiations within a world
`Simulation` in order to maximize its own total utility.

Remarks:
--------

    - When immediate_negotiations is true, negotiations start in the same step they are registered in (they may also end
      in that step) and `negotiation_speed_multiple` steps of it are conducted. That entails that requesting a
      negotiation may result in new contracts in the same time-step only of `immediate_negotiations` is true.

Simulation steps:
-----------------
    It is possible to control the order of the simulation steps differently using the `operations` parameter to the world
    constructor. this is the default order

    #. prepare custom stats (call `_pre_step_stats`)
    #. step all existing negotiations `negotiation_speed_multiple` times handling any failed negotiations and creating
       contracts for any resulting agreements
    #. Allow custom simulation (call `simulation_step` ) with stage 0
    #. run all `Entity` objects registered (i.e. all agents) in the predefined `simulation_order`.
    #. sign contracts that are to be signed at this step calling `on_contracts_finalized`  / `on_contract_signed` as
       needed
    #. execute contracts that are executable at this time-step handling any breaches
    #. allow custom simulation steps to run (call `simulation_step` ) with stage 1
    #. remove any negotiations that are completed!
    #. update basic stats
    #. update custom stats (call `_post_step_stats`)


Monitoring a simulation:
------------------------

You can monitor a running simulation using a `WorldMonitor` or `StatsMonitor` object. The former monitors events in the
world while the later monitors the statistics of the simulation.
This is a list of some of the events that can be monitored by `WorldMonitor` . World designers can add new events
either by announcing them using `announce` or as a side-effect of logging them using any of the log functions.

=================================   ===============================================================================
 Event                               Data
=================================   ===============================================================================
extra-step                           none
mechanism-creation-exception         exception: `Exception`
zero-outcomes-negotiation            caller: `Agent`, partners: `List[Agent]` , annotation: `Dict` [ `str`, `Any` ]
entity-exception                     exception: `Exception`
contract-exception                   contract: `Contract`, exception: `Exception`
agent-exception                      method: `str` , exception: `Exception`
agent-joined                         agent: `Agent`
negotiation-request                  caller: `Agent` , partners: `List` [ `Agent` ], issues: `List` [ `Issue` ]
                                     , mechanism_name: `str` , annotation: `Dict` [ `str`, `Any` ], req_id: `str`
negotiation-request-immediate        caller: `Agent` , partners: `List` [ `Agent` ], issues: `List` [ `Issue` ]
                                     , mechanism_name: `str` , annotation: `Dict` [ `str`, `Any` ]
negotiation-request-rejected         caller: `Agent` , partners: `List` [ `Agent` ] , req_id: `str`
                                     , rejectors: `List` [ `Agent` ] , annotation: `Dict` [ `str`, `Any` ]
negotiation-request-accepted         caller: `Agent` , partners: `List` [ `Agent` ] , req_id: `str`
                                     , mechanism: `Mechanism` , annotation: `Dict` [ `str`, `Any` ]
negotiation-success                  mechanism: `Mechanism` , contract: `Contract` , partners: `List` [ `Agent` ]
negotiation-failure                  mechanism: `Mechanism` , partners: `List` [ `Agent` ]
contract-executing                   contract: `Contract`
contract-nullified                   contract: `Contract`
contract-breached                    contract: `Contract`, breach: `Breach`
breach-resolved                      contract: `Contract`, breaches: `List[Breach]`, resolution
contract-executed                    contract: `Contract`
dropped-contract                     contract: `Contract`
=================================   ===============================================================================

"""
from __future__ import annotations

from .common import *
from .action import *
from .adapter import *
from .agent import *
from .entity import *
from .awi import *
from .breaches import *
from .bulletinboard import *
from .contract import *
from .helpers import *
from .mechanismfactory import *
from .mixins import *
from .monitors import *
from .save import *
from .world import *
from .simple import *
from .neg import *

__all__ = (
    common.__all__
    + action.__all__
    + adapter.__all__
    + agent.__all__
    + entity.__all__
    + awi.__all__
    + breaches.__all__
    + bulletinboard.__all__
    + contract.__all__
    + helpers.__all__
    + mechanismfactory.__all__
    + mixins.__all__
    + monitors.__all__
    + save.__all__
    + world.__all__
    + simple.__all__
    + neg.__all__
)



from __future__ import annotations

import copy
import re
import uuid
from typing import Any, Callable, Iterable

from negmas.events import Event, EventSource
from negmas.helpers.inout import ConfigReader

__all__ = ["BulletinBoard"]


class BulletinBoard(EventSource, ConfigReader):
    """
    The bulletin-board which carries all public information. It consists of sections each with a dictionary of records.
    """

    # def __getstate__(self):
    #     return self.name, self._data
    #
    # def __setstate__(self, state):
    #     name, self._data = state
    #     super().__init__(name=name)

    def __init__(self):
        """
        Constructor

        Args:
            name: BulletinBoard name
        """
        super().__init__()
        self._data: dict[str, dict[str, Any]] = {}

    def add_section(self, name: str) -> None:
        """
        Adds a section to the bulletin Board

        Args:
            name: Section name

        Returns:

        """
        self._data[name] = {}

    def query(
        self, section: str | list[str] | None, query: Any, query_keys=False
    ) -> dict[str, Any] | None:
        """
        Returns all records in the given section/sections of the bulletin-board that satisfy the query

        Args:
            section: Either a section name, a list of sections or None specifying ALL public sections (see remarks)
            query: The query which is USUALLY a dict with conditions on it when querying values and a RegExp when
            querying keys
            query_keys: Whether the query is to be applied to the keys or values.

        Returns:

            - A dictionary with key:value pairs giving all records that satisfied the given requirements.

        Remarks:

            - A public section is a section with a name that does not start with an underscore
            - If a set of sections is given, and two records in different sections had the same key, only one of them
              will be returned
            - Key queries use regular expressions and match from the beginning using the standard re.match function

        """
        if section is None:
            return self.query(
                section=[_ for _ in self._data.keys() if not _.startswith("_")],
                query=query,
                query_keys=query_keys,
            )

        if isinstance(section, Iterable) and not isinstance(section, str):
            results = [
                self.query(section=_, query=query, query_keys=query_keys)
                for _ in section
            ]
            if len(results) == 0:
                return dict()
            final: dict[str, Any] = {}
            for _ in results:
                if not _:
                    continue
                final.update(_)
            return final

        sec = self._data.get(section, None)
        if sec is None:
            return {}
        if query is None:
            return copy.deepcopy(sec)
        if query_keys:
            return {k: v for k, v in sec.items() if re.match(str(query), k) is not None}
        return {k: v for k, v in sec.items() if BulletinBoard.satisfies(v, query)}

    @classmethod
    def satisfies(cls, value: Any, query: Any) -> bool:
        method = getattr(value, "satisfies", None)
        if method is not None and isinstance(method, Callable):
            return method(query)
        if isinstance(value, dict) and isinstance(query, dict):
            for k, v in query.items():
                if value.get(k, None) != v:
                    return False
        else:
            raise ValueError(
                f"Cannot check satisfaction of {type(query)} against value {type(value)}"
            )
        return True

    def read(self, section: str, key: str) -> Any:
        """
        Reads the value associated with given key

        Args:
            section: section name
            key: key

        Returns:

            Content of that key in the bulletin-board

        """
        sec = self._data.get(section, None)
        if sec is None:
            return None
        return sec.get(key, None)

    def record(self, section: str, value: Any, key: str | None = None) -> None:
        """
        Records data in the given section of the bulletin-board

        Args:
            section: section name (can contain subsections separated by '/')
            key: The key
            value: The value

        """
        if key is None:
            try:
                skey = str(hash(value))
            except:
                skey = str(uuid.uuid4())
        else:
            skey = key
        self._data[section][skey] = value
        self.announce(
            Event("new_record", data={"section": section, "key": skey, "value": value})
        )

    def remove(
        self,
        section: list[str] | str | None,
        *,
        query: Any | None = None,
        key: str | None = None,
        query_keys: bool = False,
        value: Any = None,
    ) -> bool:
        """
        Removes a value or a set of values from the bulletin Board

        Args:
            section: The section
            query: the query to use to select what to remove
            key: the key to remove (no need to give a full query)
            query_keys: Whether to apply the query (if given) to keys or values
            value: Value to be removed

        Returns:
            bool: Success of failure
        """
        if section is None:
            return self.remove(
                section=[_ for _ in self._data.keys() if not _.startswith("_")],
                query=query,
                key=key,
                query_keys=query_keys,
            )

        if isinstance(section, Iterable) and not isinstance(section, str):
            return all(
                self.remove(section=_, query=query, key=key, query_keys=query_keys)
                for _ in section
            )

        sec = self._data.get(section, None)
        if sec is None:
            return False
        if value is not None:
            for k, v in sec.items():
                if v == value:
                    key = k
                    break
        if key is not None:
            try:
                self.announce(
                    Event(
                        "will_remove_record",
                        data={"section": sec, "key": key, "value": sec[key]},
                    )
                )
                sec.pop(key, None)
                return True
            except KeyError:
                return False

        if query is None:
            return False

        if query_keys:
            keys = [k for k in sec.keys() if re.match(str(query), k) is not None]
        else:
            keys = [k for k, v in sec.items() if v.satisfies(query)]
        if len(keys) == 0:
            return False
        for k in keys:
            self.announce(
                Event(
                    "will_remove_record",
                    data={"section": sec, "key": k, "value": sec[k]},
                )
            )
            sec.pop(k, None)
        return True

    @property
    def data(self):
        """This property is intended for use only by the world manager. No other agent is allowed to use it"""
        return self._data



from __future__ import annotations

import os
from os import walk
from pathlib import Path

import pkg_resources
import pytest

from negmas import load_genius_domain_from_folder
from negmas.inout import Scenario
from negmas.outcomes import enumerate_issues
from negmas.outcomes.outcome_space import DiscreteCartesianOutcomeSpace
from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction
from negmas.sao import AspirationNegotiator

MAX_CARDINALITY = 10_000

try:
    # try to raise the maximum limit for open files
    import resource

    resource.setrlimit(resource.RLIMIT_NOFILE, (50000, -1))
except:
    pass

GENIUSWEB_FOLDERS = Path(
    pkg_resources.resource_filename("negmas", resource_name="tests/data/geniusweb")
)


@pytest.fixture
def scenarios_folder():
    return pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/scenarios"
    )


# todo: get these to work
SCENARIOS_TO_IGNORE = [
    "S-1NIKFRT-3",
    "IntegerDomain",
    "S-1NIKFRT-2",
    "S-1NAGUNL-128",
    "S-1NAGUNL-255",
    "50issueswithRV",
    "50issues",
    "30issuesDiscounted",
    "30issueswithRV",
    "30issues",
    "50issuesDiscounted",
    "30issuesDiscountedwithRV",
    "50issuesDiscountedwithRV",
    "AgentHp2",
    "web_service",
    "four_issues",
    "AMPOvsCity",
    "laptopdomain",
    "S-1NIKFRT-1",
    "S-1NAGUNL-114",
    "laptopdomainNoBayes",
    "inheritancedomain",
    "10issues",
    "10issuesDiscounted",
    "10issueswithRV",
    "10issuesDiscountedwithRV",
    "group9-vacation",
    "group9-killer_robot",
    "FitnessA",
    "FitnessB",
    "FitnessC",
    "Grandma",
    "YXAgent",
    # "kite",
    # "group5-car_domain",
    # "group2-new_sporthal",
    # "group12-symposium",
    # "group11-car_purchase",
    # "group9-vacation",
    # "group8-holiday",
]


def get_all_scenarios():
    base = Path(__file__).parent.parent / "data" / "scenarios"
    data = []
    for root, dirs, files in walk(base):
        if len(files) == 0 or len(dirs) != 0:
            continue
        if root.split("/")[-1] in SCENARIOS_TO_IGNORE:
            continue
        data.append(root)
    return data


def test_reading_writing_linear_preferences(tmp_path):
    from negmas.preferences import LinearAdditiveUtilityFunction, UtilityFunction

    base_folder = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/Laptop"
    )
    domain = load_genius_domain_from_folder(
        base_folder,
    )
    ufuns, issues = domain.ufuns, domain.issues
    for ufun in ufuns:
        assert isinstance(ufun, LinearAdditiveUtilityFunction)
        dst = tmp_path / "tmp.xml"
        UtilityFunction.to_genius(ufun, issues=issues, file_name=dst)
        print(str(dst))
        ufun2, _ = UtilityFunction.from_genius(dst, issues=issues)
        try:
            os.unlink(dst)
        except:
            pass
        assert isinstance(ufun2, LinearAdditiveUtilityFunction)
        for outcome in enumerate_issues(issues):
            assert abs(ufun2(outcome) - ufun(outcome)) < 1e-3


def test_importing_file_without_exceptions(scenarios_folder):
    folder_name = scenarios_folder + "/other/S-1NIKFRT-1"
    load_genius_domain_from_folder(folder_name)


def test_simple_run_with_aspiration_agents():
    file_name = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/Laptop"
    )
    assert os.path.exists(file_name)
    domain = Scenario.from_genius_folder(Path(file_name))
    assert domain
    domain.to_single_issue()
    mechanism = domain.make_session(AspirationNegotiator, n_steps=100, time_limit=30)
    assert mechanism is not None
    mechanism.run()


@pytest.mark.parametrize(
    ("r0", "r1", "n_above"),
    (
        (0.0, 0.0, 11),
        (-0.1, -0.1, 11),
        (0.9, 0.9, 0),
        (0.0, 0.9, 2),
        (0.9, 0.0, 2),
        (0.0, 0.95, 1),
        (0.95, 0.0, 1),
    ),
)
def test_enumerate_discrete_rational(r0, r1, n_above):
    domain = Scenario.from_genius_folder(  # type: ignore
        Path(__file__).parent.parent
        / "data"
        / "scenarios"
        / "anac"
        / "y2013"
        / "Fifty2013",
        safe_parsing=False,
        ignore_discount=True,
    )
    assert domain
    domain.scale_max()
    domain: Scenario
    ufuns = domain.ufuns
    ufuns[0].reserved_value = r0
    ufuns[1].reserved_value = r1
    outcomes = list(domain.outcome_space.enumerate_or_sample())
    assert len(outcomes) == 11
    assert (
        len(
            list(
                domain.outcome_space.enumerate_or_sample_rational(
                    preferences=domain.ufuns, aggregator=lambda x: True  # type: ignore
                )
            )
        )
        == 0
    )
    assert (
        len(
            list(
                domain.outcome_space.enumerate_or_sample_rational(
                    preferences=domain.ufuns, aggregator=lambda x: False  # type: ignore
                )
            )
        )
        == 11
    )
    assert (
        len(
            list(
                domain.outcome_space.enumerate_or_sample_rational(
                    preferences=domain.ufuns, aggregator=any
                )
            )
        )
        == n_above
    )


def test_load_geniusweb_example_reserved_outcome():
    domain = GENIUSWEB_FOLDERS / "Fitness"
    scenario = Scenario.from_geniusweb_folder(domain, use_reserved_outcome=True)
    assert scenario is not None
    assert isinstance(scenario.outcome_space, DiscreteCartesianOutcomeSpace)
    assert len(scenario.outcome_space.issues) == 5
    for i, (name, vals) in enumerate(
        (("type", 5), ("duration", 4), ("distance", 4), ("intensity", 4), ("price", 4))
    ):
        assert scenario.outcome_space.issues[i].name == name
        assert scenario.outcome_space.issues[i].cardinality == vals
    assert scenario.outcome_space.name == "fitness"
    assert len(scenario.ufuns) == 2
    assert all(isinstance(_, LinearAdditiveUtilityFunction) for _ in scenario.ufuns)
    assert all(
        (_.reserved_value is None or _.reserved_value == float("-inf"))
        and _.reserved_outcome is not None
        for _ in scenario.ufuns
    )
    assert scenario.ufuns[0].name == "fitness1"
    assert scenario.ufuns[1].name == "fitness2"


def test_load_geniusweb_example_reserved_value():
    domain = GENIUSWEB_FOLDERS / "Fitness"
    scenario = Scenario.from_geniusweb_folder(domain, use_reserved_outcome=False)
    assert scenario is not None
    assert isinstance(scenario.outcome_space, DiscreteCartesianOutcomeSpace)
    assert len(scenario.outcome_space.issues) == 5
    for i, (name, vals) in enumerate(
        (("type", 5), ("duration", 4), ("distance", 4), ("intensity", 4), ("price", 4))
    ):
        assert scenario.outcome_space.issues[i].name == name
        assert scenario.outcome_space.issues[i].cardinality == vals
    assert scenario.outcome_space.name == "fitness"
    assert len(scenario.ufuns) == 2
    assert all(isinstance(_, LinearAdditiveUtilityFunction) for _ in scenario.ufuns)
    assert all(
        _.reserved_value is not None and _.reserved_outcome is None
        for _ in scenario.ufuns
    )
    assert scenario.ufuns[0].name == "fitness1"
    assert scenario.ufuns[1].name == "fitness2"


# @pytest.mark.parametrize("folder_name", get_all_scenarios())
# def test_importing_all_single_issue_without_exceptions(folder_name):
#     # def test_encoding_decoding_all_without_discounting():
#     # folder_name = "/Users/yasser/code/projects/negmas/tests/data/scenarios/anac/y2012/FitnessC"
#     d2 = Domain.from_genius_folder(folder_name, safe_parsing=False)
#     d2.discretize()
#     assert isinstance(d2.outcome_space, DiscreteCartesianOutcomeSpace)
#     n = d2.outcome_space.cardinality
#     if n < 10_000:
#         d2.to_single_issue()
#         assert d2.outcome_space.cardinality == n or d2.outcome_space.cardinality == float("inf")
